from __future__ import print_function

import osc
import osc.conf
import osc.core
import osc.oscerr

from functools import partial
from xml.etree import ElementTree


def valid_mpi(compiler, mpi):
    """
    It might be possible to use Intel MPI libararies and compilers from
    different Parallel Studio packages, but I currently do not want to support
    it.

    Take care to keep this in sync with the file 'macros.obs_cluster' of
    the package software:dist / mpcdf_cluster_macros
    """

    valid_intel_impi = {
        ("intel_17_0_4", "impi_2017_3"),
        ("intel_17_0_7", "impi_2017_4"),
        ("intel_18_0_1", "impi_2018_1"),
        ("intel_18_0_2", "impi_2018_2"),
        ("intel_18_0_3", "impi_2018_3"),
        ("intel_19_0_0", "impi_2019_0"),
    }

    if compiler.startswith("intel") and mpi.startswith("impi"):
        return (compiler, mpi) in valid_intel_impi
    else:
        return True


def valid_cuda(cuda, compiler):
    """
    The CUDA distribution only works with certain gcc versions,
    this little function takes care that only supported combinations
    are allowed.

    Take care to keep this in sync with the file 'macros.obs_cluster' of
    the package software:dist / mpcdf_cluster_macros
    """
    if cuda == "cuda_8_0":
        return compiler == "gcc_5"
    if cuda == "cuda_9_1":
        return compiler == "gcc_6_3_0"
    if cuda == "cuda_9_2":
        return compiler == "gcc_7"
    return False


def project_meta(api_url, project):
    return ElementTree.fromstringlist(osc.core.show_project_meta(api_url, project))


class UnsetAttributeException(Exception):
    pass


def get_attribute(api_url, project, package, attribute, with_project=False):
    attribute_meta = osc.core.show_attribute_meta(api_url, project, package, None, attribute, False, with_project)
    if attribute_meta is None:
        raise osc.oscerr.APIError("Cannot fetch value for attribute '{0}' from {1}".format(attribute, (project, package)))

    root = ElementTree.fromstringlist(attribute_meta)
    attribute = root.find("./attribute")
    if attribute is not None:
        return root
    else:
        raise UnsetAttributeException("Attribute not set")


def get_attribute_values(api_url, project, package, attribute, with_project=False):
    root = get_attribute(api_url, project, package, attribute, with_project)
    attribute = root.find("./attribute")
    return list(value.text for value in attribute.findall("./value"))


def set_attribute(api_url, project, package, attribute):
    path = ["source", project]
    if package:
        path.append(package)
    path.append("_attribute")

    attr_meta = ElementTree.tostring(attribute, encoding=osc.core.ET_ENCODING)

    url = osc.core.makeurl(api_url, path)
    resp = ElementTree.fromstringlist(osc.core.streamfile(url, osc.core.http_POST, data=attr_meta))
    if resp.find("./summary").text != "Ok":
        raise osc.oscerr.APIError("Could not store attribute")


def set_attribute_values(api_url, project, package, attribute, values):
    root = ElementTree.Element("attributes")
    attr = ElementTree.SubElement(root, "attribute")

    namespace, name = attribute.split(":")
    attr.set("namespace", namespace)
    attr.set("name", name)

    for value in values:
        v = ElementTree.SubElement(attr, "value")
        v.text = value

    set_attribute(api_url, project, package, root)


def has_attribute(api_url, project, package, attribute):
    path = ["source", project]
    if package:
        path.append(package)
    path.append("_attribute")
    path.append(attribute)

    url = osc.core.makeurl(api_url, path)
    resp = ElementTree.fromstringlist(osc.core.streamfile(url, osc.core.http_GET))

    namespace, name = attribute.split(":")

    for a in resp.findall("./attribute"):
        if a.get("namespace") == namespace and a.get("name") == name:
            return True
    return False


def remove_attribute(api_url, project, package, attribute_name):
    path = ["source"] + [project]
    if package:
        path.append(package)
    path.append("_attribute")
    path.append(attribute_name)

    url = osc.core.makeurl(api_url, path)
    resp = ElementTree.fromstringlist(osc.core.streamfile(url, osc.core.http_DELETE))
    if resp.find("./summary").text != "Ok":
        raise osc.oscerr.APIError("Could not remove attribute")


def mpcdf_enable_repositories(api_url, project, package, verbose=False):
    from itertools import product
    import sys

    root = ElementTree.fromstringlist(osc.core.show_package_meta(api_url, project, package))
    build = root.find("./build")
    if build is None:
        build = ElementTree.SubElement(root, "build")
    else:
        for enable in build.findall("./enable"):
            build.remove(enable)

    try:
        enable_repos = get_attribute_values(api_url, project, package, "MPCDF:enable_repositories")
    except Exception:
        if verbose:
            print("Warning: Could not get attribute MPCDF:enable_repositories for package {0}, skipping".format(package))
        return False

    def try_get_attribute(package, attribute, with_project=False):
        try:
            return get_attribute_values(api_url, project, package, "MPCDF:" + attribute, with_project=with_project)
        except UnsetAttributeException:
            print("ERROR: Attribute MPCDF:" + attribute + " is not set, aborting here", file=sys.stderr)
            raise SystemExit(1)

    compilers = try_get_attribute(package, "compiler_modules", with_project=True)
    all_compilers = try_get_attribute(None, "compiler_modules")
    mpis = try_get_attribute(package, "mpi_modules", with_project=True)
    cudas = try_get_attribute(package, "cuda_modules", with_project=True)

    default_compilers = try_get_attribute(None, "default_compiler")
    default_mpis = try_get_attribute(None, "default_mpi")
    default_cudas = try_get_attribute(None, "default_cuda")

    def enable(name):
        node = ElementTree.Element("enable")
        node.set("repository", name)
        node.tail = "\n    "
        build.insert(0, node)
        if verbose:
            print("Enabling", name)

    def actual_compilers():
        for compiler in compilers:
            if compiler == "intel":
                for intel_compiler in filter(lambda cc: cc.startswith("intel"), all_compilers):
                    yield intel_compiler
            elif compiler == "gcc":
                for gcc_compiler in filter(lambda cc: cc.startswith("gcc"), all_compilers):
                    yield gcc_compiler
            elif compiler == "default_compiler":
                for default_compiler in default_compilers:
                    yield default_compiler
            else:
                yield compiler

    def actual_mpis():
        for mpi in mpis:
            if mpi == "impi":
                for impi in filter(lambda cc: cc.startswith("impi"), mpis):
                    yield impi
            elif mpi == "default_mpi":
                for default_mpi in default_mpis:
                    yield default_mpi
            else:
                yield mpi

    def actual_cudas():
        for cuda in cudas:
            if cuda == "default_cuda":
                for default_cuda in default_cudas:
                    yield default_cuda
            else:
                yield cuda

    for flag in enable_repos:

        if flag == "system":
            enable("System")

        if flag == "compilers":
            for compiler in actual_compilers():
                enable(compiler)

        if flag == "mpi":
            for mpi, compiler in product(actual_mpis(), actual_compilers()):
                if valid_mpi(compiler, mpi):
                    enable(mpi + "_" + compiler)

        if flag == "cuda":
            for cuda, compiler in product(actual_cudas(), all_compilers):
                if valid_cuda(cuda, compiler):
                    enable(cuda + "_" + compiler)

        if flag == "cuda_mpi":
            for cuda, mpi, compiler in product(actual_cudas(), actual_mpis(), all_compilers):
                if valid_cuda(cuda, compiler) and valid_mpi(compiler, mpi):
                        enable(cuda + "_" + mpi + "_" + compiler)

    if len(build.getchildren()) > 0:
        build.getchildren()[-1].tail = "\n  "

    pkg_meta = ElementTree.tostring(root, encoding=osc.core.ET_ENCODING)

    osc.core.edit_meta("pkg", (project, package), data=pkg_meta)
    return True


def mpcdf_setup_repositories(api_url, project, distribution=None, parent=None, packages=None, dry_run=False):
    if parent:
        for attribute in ["compiler_modules", "default_compiler", "mpi_modules", "default_mpi", "cuda_modules", "default_cuda"]:
            print("Copying attribute 'MPCDF:{0}' from parent project".format(attribute))
            set_attribute(api_url, project, None, get_attribute(api_url, parent, None, "MPCDF:" + attribute))

    compilers = list(get_attribute_values(api_url, project, None, "MPCDF:compiler_modules"))
    mpis = list(get_attribute_values(api_url, project, None, "MPCDF:mpi_modules"))
    cudas = list(get_attribute_values(api_url, project, None, "MPCDF:cuda_modules"))

    if distribution is None:
        # Get existing value from project meta
        dist_repo = project_meta(api_url, project if parent is None else parent).find(
            "./repository[@name='System']/path[@project='distributions']")
        if dist_repo is not None:
            distribution = dist_repo.get("repository")
        else:
            raise osc.oscerr.WrongArgs("ERROR: Specify distribution or parent project")
        print("Using '{0}' as base distribution".format(distribution))

    distributions = project_meta(api_url, "distributions")
    dist_repo = distributions.find('./repository[@name="{0}"]'.format(distribution))
    if dist_repo is None:
        raise osc.oscerr.WrongArgs("No repository '{0}' is defined in project 'distributions' on the server".format(distribution))
    architectures = list(arch.text for arch in dist_repo.findall("./arch"))

    root = project_meta(api_url, project)

    prjconf = osc.core.show_project_conf(api_url, project)
    start_marker = "# Autogenerated by osc mpcdf_setup_repos, do not edit till end of section\n"
    end_marker = "# End of autogenerated section\n"

    try:
        start = prjconf.index(start_marker)
        end = prjconf.index(end_marker)
    except ValueError:
        start = None
        end = len(prjconf)

    prjconf_head = "".join(prjconf[:start])
    prjconf_tail = "".join(prjconf[end + 1:])
    prjconf = [start_marker]

    # Remove existing repositories
    for oldrepo in root.findall("./repository"):
        root.remove(oldrepo)

    def repo(name, *dependencies, **kwargs):
        is_compiler = kwargs.pop("compiler", False)
        is_mpi = kwargs.pop("mpi", False)
        is_cuda = kwargs.pop("cuda", False)
        is_cuda_mpi = kwargs.pop("cuda_mpi", False)

        cuda_repo = kwargs.pop("cuda_repo", "")

        have_compiler = is_compiler or is_mpi or is_cuda or is_cuda_mpi
        have_mpi = is_mpi or is_cuda_mpi
        have_cuda = is_cuda or is_cuda_mpi

        r = ElementTree.SubElement(root, "repository")
        r.set("name", name)
        r.text = "\n    "

        def path(project, repo):
            p = ElementTree.SubElement(r, "path")
            p.set("project", project)
            p.set("repository", repo)
            p.tail = "\n    "

        if parent:
            path(parent, name)
        for dep_project, dep_repo in dependencies:
            path(dep_project, dep_repo)

        for arch in architectures:
            a = ElementTree.SubElement(r, "arch")
            a.text = arch
            a.tail = "\n    "
        a.tail = "\n  "
        r.tail = "\n  "

        prjconf.append("%if %_repository == {0}".format(name))
        prjconf.append("Macros:")

        prjconf.append("%is_compiler_repository {0}".format(1 if is_compiler else 0))
        prjconf.append("%is_mpi_repository {0}".format(1 if is_mpi else 0))
        prjconf.append("%is_cuda_repository {0}".format(1 if is_cuda else 0))
        prjconf.append("%is_cuda_mpi_repository {0}".format(1 if is_cuda_mpi else 0))

        prjconf.append("%have_mpcdf_compiler {0}".format(1 if have_compiler else 0))
        prjconf.append("%have_mpcdf_mpi {0}".format(1 if have_mpi else 0))
        prjconf.append("%have_mpcdf_cuda {0}".format(1 if have_cuda else 0))

        if is_cuda:
            prjconf.append("%cuda_repository {0}".format(cuda_repo))

        for macro, value in kwargs.items():
            prjconf.append("%{0} {1}".format(macro, value))

        prjconf.append(":Macros")
        prjconf.append("%endif")
        prjconf.append("")

    if parent:
        repo("System")
    else:
        repo("System", ("distributions", distribution))

    for compiler in compilers:
        repo(compiler, (project, "System"), compiler=True)

        for mpi in filter(partial(valid_mpi, compiler), mpis):
            repo(mpi + "_" + compiler, (project, compiler), mpi=True, mpi_repository=mpi)

    for cuda in cudas:
        for compiler in filter(partial(valid_cuda, cuda), compilers):
            repo(cuda + "_" + compiler, (project, compiler), cuda=True, cuda_repo=cuda)
            for mpi in filter(partial(valid_mpi, compiler), mpis):
                repo(cuda + "_" + mpi + "_" + compiler,
                     (project, cuda + "_" + compiler),
                     (project, mpi + "_" + compiler),
                     cuda_mpi=True)

    # Remove build configuration
    build = root.find("./build")
    if build is None:
        build = ElementTree.Element("build")
        build.text = "\n    "
        disable = ElementTree.SubElement(build, "disable")
        disable.tail = "\n  "
        build.tail = "\n  "
        root.insert(list(root).index(root.find("./repository")), build)

    root.getchildren()[-1].tail = "\n"
    prj = ElementTree.tostring(root, encoding=osc.core.ET_ENCODING)

    prjconf.append(end_marker)
    prjconf = prjconf_head + "\n".join(prjconf) + prjconf_tail

    if dry_run:
        print("osc meta prj {0} -F - <<EOF\n{1}\nEOF\n".format(project, prj))
        print("osc meta prjconf {0} -F - <<EOF\n{1}\nEOF\n".format(project, prjconf))
    else:
        # First set-up the <enable/> flags, that way no
        # spurious builds are launched
        if packages is None:
            packages = osc.core.meta_get_packagelist(api_url, project)
        for package in packages:
            print("Updating repositories for", package)
            if not mpcdf_enable_repositories(api_url, project, package):
                print("ATTENTION: Not changing unmanaged package {0}".format(package))

        # Update repositories
        print("Updating prj meta")
        osc.core.edit_meta("prj", project, data=prj, force=True)
        print("Updating prjconf meta")
        osc.core.edit_meta("prjconf", project, data=prjconf)


def set_as_branch(api_url, my_project, my_package, main_project, main_package):
    import os

    def dirmeta(project, package):
        url = osc.core.makeurl(api_url, ["source", project, package])
        return ElementTree.fromstringlist(osc.core.streamfile(url, osc.core.http_GET))

    srcmeta = dirmeta(my_project, my_package)
    dstmeta = dirmeta(main_project, main_package)

    linkinfo = srcmeta.find('./linkinfo')
    if linkinfo is not None:
        print("ERROR: package {0} is already linked to package {1} in project {2}".format(
              my_package, linkinfo.get("package"), linkinfo.get("project")))
        return False

    linkfile = ElementTree.Element("link")
    linkfile.set("project", main_project)
    linkfile.set("package", main_package)
    linkfile.set("baserev", dstmeta.get("srcmd5"))
    patches = ElementTree.SubElement(linkfile, "patches")
    ElementTree.SubElement(patches, "branch")

    url = osc.core.makeurl(api_url, ["source", my_project, my_package, "_link"])
    resp = ElementTree.fromstringlist(osc.core.streamfile(url, osc.core.http_PUT, data=ElementTree.tostring(linkfile)))
    rev = resp.get("rev")
    if rev is None:
        print("ERROR: Could not commit _link file")
        return False
    else:
        print("Commited as revision", rev)
        this_package = osc.core.store_read_package(os.curdir)
        this_project = osc.core.store_read_project(os.curdir)
        if this_package == my_package and this_project == my_project:
            pac = osc.core.filedir_to_pac(os.curdir)
            rev = pac.latest_rev(expand=True)
            pac.update(rev)
    return True
